"""This module contains the config parser functions."""

import os

import confipy.reader
import confipy.converter

ERR_CFG_NOT_FOUND = "Cannot find referenced config '{}'."


def parsing_handler(parsers, cfg, source_path=None, splitter=" + ",
                    marker="$", **kwargs):
    """Manages processing of all parsers. Handles relevant arguments and
    keyword arguments to correct parser functions.

    Returns flattened, parsed config data.

    """

    # parser_name: (func, args, kwargs)
    parsers_opt = {"include": (include, [source_path], {}),
                   "substitute": (substitute, [], {"splitter": splitter,
                                                   "marker": marker})}

    parsed_cfg = confipy.converter._flat_dict(cfg)
    for parser in parsers:
        func, args, kwargs = parsers_opt[parser]
        parsed_cfg = func(parsed_cfg, *args, **kwargs)

    return parsed_cfg


def include(flattened_dict, source_path, marker="$include"):
    """Scan config dictionary for include statements. Load and insert
    referenced config files under corresponding key's namespace.

    Parameters
    ----------
    flattened_dict: dict
        Flattened dictionary containing config data.
    source_path: str
        Path to original config file.

    Return
    ------
    parsed_dict: dict

    """

    cwd = os.path.dirname(source_path)
    parsed_dict = {}

    for key_chain, value in flattened_dict.items():
        if not value.startswith(marker):
            parsed_dict[key_chain] = value
            continue

        absolute_path = value.replace(marker, "").lstrip().rstrip()
        relative_path = os.path.join(cwd, absolute_path)

        if os.path.exists(relative_path):
            path = relative_path
        elif os.path.exists(absolute_path):
            path = absolute_path
        else:
            raise AssertionError(ERR_CFG_NOT_FOUND.format(absolute_path))

        inc_cfg = confipy.reader.read_config(path)
        inc_flat = confipy.converter._flat_dict(inc_cfg, list(key_chain))
        inc_included = include(inc_flat, path)
        parsed_dict.update(inc_included)

    return parsed_dict


def substitute(to_parse, parsed=None, splitter=" + ", marker="$"):
    """Find keys identified by splitter and marker signs. Only values
    containing the splitter are taken into account. Splitted values must have
    keys which begin with the marker sign. Otherwise, keys are ignored.

    The function calls itself recursively until to_parse is empty.

    Parameters
    ----------
    to_parse: dict
        Dictionary with items to be parsed.
    parsed: dict
        Dictionary with valid lookup items for substitution usage.
    splitter: str
        The splitter to identify possible keys.
    marker: str
        The marker to identify correct keys.

    Return
    ------
    parsed: dict

    """

    # return result if to_parse is empty
    if not to_parse:
        return parsed

    # if there's nothing yet, find all non-subs items
    if not parsed:
        parsed = {key: value for key, value in to_parse.items()
                  if not _contains(value, splitter)}
        to_parse = {key: value for key, value in to_parse.items()
                    if key in set(parsed) ^ set(to_parse)}

    # iterate items to be parsed; distinguish strings and lists
    for key, value in to_parse.copy().items():
        if isinstance(value, str):
            valid = _substitue_value(value, parsed, splitter, marker)
        else:
            valid = [_substitue_value(element, parsed, splitter, marker)
                     if _contains(element, splitter) else element
                     for element in value]

        if all(valid):
            del to_parse[key]
            parsed[key] = valid

    return substitute(to_parse, parsed)


def _substitue_value(value, parsed, splitter, marker):
    """Substitute keys of a singular string. Returns False if one key could
    not successfully be replaced.

    Parameters
    ----------
    value: str
        String value containing keys to be substituted.
    parsed: dict
        Dictionary with valid lookup items for substitution usage.
    splitter: str
        The splitter to identify possible keys.
    marker: str
        The marker to identify correct keys.

    Return
    ------
    parsed_item: str, None

    """

    parsed_item = ""
    for part in value.split(splitter):
        if not part.startswith(marker):
            parsed_item += part
            continue

        key_chain = _convert_key_chain(part[1:])
        parsed_value = parsed.get(key_chain)

        if not parsed_value:
            return False

        parsed_item += parsed_value

    return parsed_item


def _contains(values, splitter):
    """Check presence of marker in values.

    Parameters
    ----------
    values: str, iterable
        Either a single value or a list of values.
    splitter: str
        The target to be searched for.

    Return
    ------
    boolean

    """

    if isinstance(values, str):
        values = (values,)

    if any([splitter in x for x in values]):
        return True
    else:
        return False


def _convert_key_chain(key_string):
    """Convert dot notation to tupled key chain notation"""

    return tuple(key_string.split("."))



"""This module provides the index function."""

import confipy.reader
import confipy.converter
import confipy.parser

converter_opts = {"dot": confipy.converter.DotNotation,
                  "dict": dict}


def load(path_or_fp, read_engine="auto", parsers=("include", "substitute"),
         converter="dot", **kwargs):
    """Main function to initiate reading, parsing and conversion.

    Parameters
    ----------
    path_or_fp: str, file-like
        Path to config file to be read.
    read_engine: {"auto", "yaml", "json", "configparser"}, optional
        Define the read engine to open the config file. By default, the file
        type is used to infer the correct read engine.
    parsers: iterable, optional
        Define parsers to be run on raw config data. Be aware, order matters.
    converter: {"dot", "dict"}, optional
        Define output type of config data. By default, config data provided in
        dot notation.

    Returns
    -------
    converted_cfg: {confipy.converter.DotNotation, dict}

    """

    read_cfg = confipy.reader.read_config(path_or_fp, read_engine=read_engine)
    parsed_cfg = confipy.parser.parsing_handler(parsers, read_cfg,
                                                source_path=path_or_fp,
                                                **kwargs)

    conv_type = converter_opts[converter]
    converted_cfg = confipy.converter._unflat_dict(parsed_cfg,
                                                   default_type=conv_type)

    return converted_cfg



"""This module contains converter functions."""

class DotNotation(object):
    """Enables dot notation for accessing config properties."""

    def get(self, key, default=None):
        if hasattr(self, key):
            return getattr(self, key)
        return default

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __repr__(self):
        tpl = "CfgNode: {} nodes ({}) / {} values ({})"
        nodes = [x for x in vars(self)
                 if isinstance(getattr(self,x), DotNotation)]
        values = [x for x in vars(self)
                  if not isinstance(getattr(self,x), DotNotation)]
        return tpl.format(len(nodes), nodes, len(values), values)


def _unflat_dict(flat_dict, unflat_dict=None, default_type=dict):
    """Convert flattened dict back to nested dict structure.

    Parameters
    ----------
    flat_dict: dict
        See _flatten_dict() for more information.
    unflat_dict: None, dict
        Parameter is used as parent dictionary.

    Return
    ------
    unflattend_dict: dict

    """

    if not unflat_dict:
        unflat_dict = default_type()

    # iterate, begin with lowest depth
    sorted_items = sorted(flat_dict.items(), key=lambda x: len(x[0]))
    for key_chain, value in sorted_items:
        this_key = key_chain[0]
        # directly set value, if only one key is present
        if len(key_chain)== 1:
            unflat_dict[this_key] = value
            continue

        # if key exists, do not create again
        key_exists = unflat_dict.get(this_key)
        if key_exists:
            _unflat_dict({key_chain[1:]: value}, key_exists, default_type)
            continue

        unflat_dict[this_key] = _unflat_dict({key_chain[1:]: value},
                                             default_type=default_type)

    return unflat_dict


def _flat_dict(cfg_dict, parent=None):
    """Convert nested dictionary into flattened dict of key-chain
    value pairs where the key-chain is a tuple of nested keys.

    For example, {"key1": {"key11":{"key111": "value1"}}}
    evaluates to  {("key1", "key11", "key111"): "value1"}.

    Simplifies dictionary operations for other parser functions.

    Parameters
    ----------
    cfg_dict: dict
        Dictionary containing config data.
    parent: None, optional
        For nested, recursive calls remembers the parent of current level.

    Return
    ------
    flattened: dict

    """

    if not parent:
        parent = []

    flattened = {}
    for key, value in cfg_dict.items():
        key_level = parent + [key]
        if isinstance(value, dict):
            flattened.update(_flat_dict(value, key_level))
        else:
            key_chain = tuple(key_level)
            flattened[key_chain] = value

    return flattened


from setuptools import setup, find_packages

setup(
    name='confipy',
    version='0.0.1',
    packages=find_packages(),
    url='https://github.com/mansenfranzen/confipy',
    license='MIT',
    author='Franz Woellert',
    author_email='franz.woellert@gmail.com',
    description='A convenient config file reader.',
    install_requires=['PyYAML'],
    download_url = 'https://github.com/mansenfranzen/confipy/tarball/0.0.1',
    keywords = ['config', 'yaml']
)



"""This module contains the config reader."""

import yaml
import json
import collections

# python 2.7
try:
    import configparser
except ImportError:
    import ConfigParser as configparser


def _read_configparser(fp):
    """Read configparser compliant config file from file like object.

    Parameters
    ----------
    fp: file-like object
        Any object having a read() method.

    Return
    ------
    cfg_dict: dict
        Dictionary containing config data.

    """

    cfg_parser = configparser.ConfigParser()

    # python 2.7
    try:
        cfg_parser.read_file(fp)
    except AttributeError:
        cfg_parser.readfp(fp)

    cfg_dict = {section: dict(cfg_parser.items(section))
                for section in cfg_parser.sections()}
    return cfg_dict


READER = collections.OrderedDict({"yaml": yaml.load,
                                  "configparser": _read_configparser,
                                  "json": json.load})

MIME_TYPES = {"yaml": READER["yaml"],
              "ini": READER["configparser"],
              "cfg": READER["configparser"],
              "json": READER["json"]}


def _read_infer(fp):
    """Iterates config reader functions and tries to find correct reader via
    try except statements.

    Parameters
    ----------
    fp: file-like object
        Any object having a read() method.

    Return
    ------
    cfg_dict: dict
        Dictionary containing config data.

    Raises
    ------
    IOError if no reader functions applies.

    """

    for read_label, read_func in READER.items():
        try:
            cfg_dict = read_func(fp)
        except Exception as error:
            print("Reader '{}' fails.".format(read_label))
        else:
            print("Reader '{}' succeeds.".format(read_label))
            return cfg_dict
        finally:
            fp.seek(0)
    else:
        raise IOError("Cannot find appropriate reader.")


def read_config(path_or_fp, read_engine="auto"):
    """Read config file and return as dictionary.

    Parameters
    ----------
    path_or_fp: str, file-like
        Path to config file to be read.
    read_engine: {"auto", "yaml", "json", "configparser"}
        Define the read engine to open the config file. By default, the file
        type is used to infer the correct read engine.

    Return
    ------
    cfg_dict: dict
        Key value pairs of config file.

    """

    # check for path or file-like object
    if hasattr(path_or_fp, "read"):
        fp = path_or_fp
        mime = read_engine
    else:
        fp = open(path_or_fp, "r")
        mime = path_or_fp.split(".")[-1].lower()

    read_function = MIME_TYPES.get(mime, _read_infer)
    cfg_dict = read_function(fp)

    return cfg_dict
